#!/usr/bin/perl

if (@ARGV == 0) {
  print STDERR "Usage:  ./chasm.pl source.asm\r\n";
  print STDERR "Output file is always avr.bin\r\n";
  exit 1;
};

open FILE, "instructions.txt";
  while ($line = <FILE>) {
    chomp $line;
    $line = lc($line);
    ($code, $text) = $line =~ /([^ ]+) (.*)/;
    $code{$text} = reverse pack('H*', $code);
  };
close FILE;

loadsource($ARGV[0]);

sub loadsource {
  my $FILE;
  my ($line, $file, $ln, @part, $p);
  open($FILE, '<', $_[0]) or die "Unable to open $_[0] for reading...\n";
    while ($line = <$FILE>) {
      chomp $line; $ln++;
      $line =~ s/ *\#.*//g;
      $line =~ s/^ +//g;
      $line =~ s/ +$//g;
      $line =~ s/  +/ /g;
      if ($line =~ /^include /) {
        ($file) = $line =~ /include ([^ ]+)/;
        loadsource($file);
      } else {
        $line = lc($line);
        @part = split /\;/, $line;
        for ($p = 0; $p < @part; $p++) {
          $part[$p] =~ s/^ +//g;
          $part[$p] =~ s/ +$//g;
          push @source, $part[$p];
          push @where, "$_[0]:$ln";
          if ($p > 0) {
            $where[-1] .= "." . ($p + 1);
          };
        };
      };
    };
  close $FILE;
};

$ram_base = -1;
$ram_size = -1;
for ($pass = 0; $pass < 2; $pass++) {
  $last_label = '';
  $memory = ''; $ip = 0;
  %replace = ();

#  $debug = 1;

  for ($ln = 0; $ln < @source; $ln++) {
    $line = $source[$ln];
    if (@macro) {
      $ln--;
      $line = shift @macro;
    };
    print "New Line: $line\n" if $debug;
    next if $line eq '';
    if ($line =~ /^[a-z][a-z0-9\_]+\:$/ or $line =~ /^\.[a-z0-9\_]+$/) {
      ($label) = $line =~ /([\.a-z][a-z0-9\_]*)/;
      if ($label =~ /^\./) {
        $address{"$last_label$label"} = $ip;
      } else {
        $address{"$label"} = $ip;
        $last_label = $label;
      };
    } else {
      ($opcode) = $line =~ /([^ ]+)/;
      if ($opcode eq 'replace') {
        ($search, $replace) = $line =~ /replace ([^ ]+) ([^ ]+)/;
        $replace{$search} = $replace;
      } elsif ($opcode eq 'goto') {
        $line =~ s/(\$[0-9a-f]+)/hex(substr($1, 1))/ge;
        $line =~ s/(0x[0-9a-f]+)/hex(substr($1, 2))/ge;
        ($ip) = $line =~ / (\d+)/;
        $memory .= ("\x00" x (2 * $ip - length($memory))) if length($memory) < 2 * $ip;
      } elsif ($opcode eq 'device_type') {
        $line =~ s/(\$[0-9a-f]+)/hex(substr($1, 1))/ge;
        $line =~ s/(0x[0-9a-f]+)/hex(substr($1, 2))/ge;
        ($config[0]) = $line =~ / (\d+)/;
      } elsif ($opcode eq 'fuse_low') {
        $line =~ s/(\$[0-9a-f]+)/hex(substr($1, 1))/ge;
        $line =~ s/(0x[0-9a-f]+)/hex(substr($1, 2))/ge;
        ($config[1]) = $line =~ / (\d+)/;
      } elsif ($opcode eq 'fuse_high') {
        $line =~ s/(\$[0-9a-f]+)/hex(substr($1, 1))/ge;
        $line =~ s/(0x[0-9a-f]+)/hex(substr($1, 2))/ge;
        ($config[2]) = $line =~ / (\d+)/;
      } elsif ($opcode eq 'fuse_extended') {
        $line =~ s/(\$[0-9a-f]+)/hex(substr($1, 1))/ge;
        $line =~ s/(0x[0-9a-f]+)/hex(substr($1, 2))/ge;
        ($config[3]) = $line =~ / (\d+)/;
      } elsif ($opcode eq 'sram') {
        $line =~ s/(\$[0-9a-f]+)/hex(substr($1, 1))/ge;
        $line =~ s/(0x[0-9a-f]+)/hex(substr($1, 2))/ge;
        ($ram_base, $ram_size) = $line =~ /\s+(\d+)\s+(\d+)/;
      } elsif ($opcode eq 'allocate') {
        $line =~ s/(\$[0-9a-f]+)/hex(substr($1, 1))/ge;
        $line =~ s/(0x[0-9a-f]+)/hex(substr($1, 2))/ge;
        ($trash, $name, $offset, $size) = split /\s+/, $line;
        if ($size eq '') {
          $size = $offset;
          $offset = -1;
        };
        if ($ram_size == -1) {
          print STDERR "$where[$ln]: Must use 'sram' directive first.\n"; $fatal = 1;
        } elsif (($pass == 0 and $offset >= 0) or ($pass == 1 and $offset < 0)) {
          if ($allocations eq '') {
            $allocations = '1' x $ram_base;
            $allocations .= '0' x $ram_size;
          };
          $string = '0' x $size;
          if ($offset == -1) {
            $offset = index($allocations, $string);
          } else {
            if (substr($allocations, $offset, $size) ne $string) {
              $offset = -1;
            };
          };
          if ($offset == -1) {
            print STDERR "$where[$ln]: Cannot make this allocation.\n"; $fatal = 1;
          } else {
            $string = '1' x $size;
            substr($allocations, $offset, $size) = $string;
            $address{$name} = $offset;
            $address{"${name}_lsb"} = $offset & 255;
            $address{"${name}_msb"} = $offset >> 8;
            print "Allocated '$name' at offset $offset with size $size. (ram_base = $ram_base, ram_size = $ram_size)\n" if $debug;
          };
        };
      } elsif ($opcode eq 'delay') {
        ($clocks) = $line =~ /delay ([0-9]+)/;
        if ($clocks == 0) {
        } elsif ($clocks == 1) {
          @macro = ('nop');
        } elsif ($clocks == 2) {
          @macro = ('nop2');
        } elsif ($clocks == 3) {
          @macro = ('nop', 'nop2');
        } elsif ($clocks == 4) {
          @macro = ('nop2', 'nop2');
        } elsif ($clocks == 5) {
          @macro = ('nop', 'nop2', 'nop2');
        } elsif ($clocks == 6) {
          @macro = ('nop2', 'nop2', 'nop2');
        } elsif ($clocks >= 7 and $clocks < 10) {
          if (!$added_delay_calls) {
            push @source, ('delay9:', 'nop', 'delay8:', 'nop', 'delay7:', 'ret');
            $added_delay_calls = 1;
          };
          @macro = ("call delay$clocks");
        } elsif ($clocks >= 10) {
          if (!$added_delay_loop) {
            push @source, ('delay_loop:', 'dec r23', 'jic z delay_loop', 'ret');
            $added_delay_loop = 1;
          };
          if ($clocks > 777) {
            @macro = ('ld r23 253', 'call delay_loop');
            $clocks -= 7 + 3 * 253;
          };
          # 10 <= clocks <= 777
          $clocks -= 7;
          $cycles = int($clocks / 3);
          if ($cycles > 256) {
            print STDERR "$where[$ln]: Can't delay this long!\n"; $fatal = 1;
          } elsif ($cycles == 256) {
            $cycles = 0;
          };
          $clocks -= 3 * $cycles;
          @macro = ("ld r23 $cycles", "call delay_loop");
          push @macro, 'nop2' if $clocks == 2;
          push @macro, 'nop' if $clocks == 1;
        };
      } elsif ($opcode eq 'words') {
        $line =~ s/(\$[0-9a-f]+)/hex(substr($1, 1))/ge;
        $line =~ s/(0x[0-9a-f]+)/hex(substr($1, 2))/ge;
        @list = split /\s+/, $line;
        shift @list;
        while (@list) {
          $item = shift @list;
          substr($memory, 2 * $ip, 2) = pack('v', $item); $ip++;
        };
      } else {
        while ('true') {
          $change = 0;
          foreach $item (keys %replace) {
            $original = $line;
            $line =~ s/\b$item\b/$replace{$item}/g;
            $change++ if $line ne $original;
          };
          last unless $change;
        };
        foreach $item (keys %replace) {
          $line =~ s/\b$item\b/$replace{$item}/g;
        };
        print "Line (1): $line\n" if $debug;
        $line =~ s/(\$[0-9a-f]+)/hex(substr($1, 1))/ge;
        $line =~ s/(0x[0-9a-f]+)/hex(substr($1, 2))/ge;
        print "Line (2): $line\n" if $debug;
        if ($opcode eq 'call' or $opcode eq 'jmp' or $opcode eq 'jp' or $opcode eq 'jis' or $opcode eq 'jic') {
          if ($pass == 0) {
            substr($memory, 2 * $ip, 2) = "\x00\x00"; $ip++;
          } else {
            ($label) = $line =~ /([\.a-z][a-z0-9\_]*)$/;
            if ($label =~ /^\./) {
              $look = "$last_label$label";
            } else {
              $look = $label;
            };
            if (exists $address{$look}) {
              $offset = $address{$look} - $ip - 1;
              $search = $line;
              $search =~ s/$label/$offset/g;
              if (exists $code{$search}) {
                substr($memory, 2 * $ip, 2) = $code{$search}; $ip++;
              } else {
                print STDERR "$where[$ln]: Can't find '$search' in instruction list.\n"; $fatal = 1;
              };
            } else {
              print STDERR "$where[$ln]: Label '$label' does not exist.\n"; $fatal = 1;
              $" = "', '";
              @labels = keys %address;
              print "Labels: '@labels'\n" if $debug;
            };
          };
        } else {
          $fun = ' ' . $line . ' ';
          foreach $item (split / /, $line) {
            next if $item eq '';
            if (exists $address{$item}) {
              $fun =~ s/ $item / $address{$item} /g;
            };
          };
          $line = substr($fun, 1, -1);
          $address = '';
          TRY_AGAIN:
          $search = $line;
          if (exists $code{$search}) {
            #print "$line\n";
            substr($memory, 2 * $ip, 2) = $code{$search}; $ip++;
            if ($address ne '') {
              substr($memory, 2 * $ip, 2) = pack('v', $address); $ip++;
            };
          } else {
            if ($line =~ /in .+ \[/) {
              $line =~ s/^in/ld/;
              goto TRY_AGAIN;
            };
            if ($line =~ /out \[/) {
              $line =~ s/^out/ld/;
              goto TRY_AGAIN;
            };
            if ($line =~ /ld (.+ )?\[[0-9]+\]/ and $address eq '') {
              ($address) = $line =~ /\[(.*?)\]/;
              $line =~ s/\[.*?\]/\[\_\_\_\_\]/;
              goto TRY_AGAIN;
            };
            print STDERR "$where[$ln]: Can't find '$search' in instruction list.\n"; $fatal = 1;
          };
        };
      };
    };
  };

  if ($fatal) {
    print STDERR "It's $ip bytes so far...\n";
    exit 1;
  };

};

open FILE, '>', "avr.bin";
  print FILE $memory;
close FILE;
open FILE, '>', "avr.bs";
  print FILE pack('vCCC', @config);
close FILE;

print STDERR "Wrote $ip bytes to avr.bin!\n";
