include attiny48.asm
#fuse_low $EF  # Low-speed clock...

replace usb_rd pbout.0
replace usb_wr pbout.1
replace latch_wr pbout.2
replace usb_re pbin.6
replace usb_te pbin.7
replace data_in pdin
replace data_out pdout
replace data_dir pddir

replace gal_data_in pcout.0
replace gal_clock pcout.1
replace gal_strobe pcout.2
replace gal_program pcout.3
replace gal_data_out pcout.4

ld r16 $07; out pbout r16; ld r16 $07; out pbdir r16
ld r16 $00; out pcout r16; ld r16 $0F; out pcdir r16
ld r16 $00; call latch_write

lock:
  call usb_read; cmp r16 $83; jic z lock
  call usb_read; cmp r16 $84; jic z lock
  call usb_read; cmp r16 $D4; jic z lock
  call usb_read; cmp r16 $F8; jic z lock
  call usb_read; cmp r16 $21; jic z lock
  call usb_read; cmp r16 $D5; jic z lock
  call usb_read; cmp r16 $58; jic z lock
  call usb_read; cmp r16 $6B; jic z lock

  ld r16 $40; call latch_write      # Activate 5v power input...
  call delay5ms
  ld r16 $06; out pcout r16         # Raise strobe and clock lines...
  call delay5ms
  ld r16 $C0; call latch_write      # Activate 12v programming voltage...
  call delay5ms

loop:
  ld r16 16; .1; dec r16; jic z .1
  call usb_read
  ld r17 r16; and r17 $80; jic z shutdown
  ld r17 r16; and r17 $40; jic z set_bits
  ld r17 r16; and r17 $20; jic z delay
  ld r17 r16; and r17 $0F; out pcout r17
  ld r17 r16; and r17 $10; jis z loop
  in r16 pcin; and r16 $10
  shr r16; shr r16; shr r16; shr r16
  or r16 $30; call usb_write
jmp loop

delay:
  call delay5ms
jmp loop

set_bits:
  or r16 $C0; call latch_write
jmp loop

shutdown:
  ld r16 $00; out pcout r16
  ld r16 $40; call latch_write
  call delay5ms
  ld r16 $00; call latch_write
  call delay5ms
jmp lock

usb_read:
  sic usb_re
  jmp usb_read
  clr usb_rd
  nop
  in r16 data_in
  set usb_rd
ret

usb_write:
  push r17
  sic usb_te
  jmp usb_write
  out data_out r16
  ld r17 $FF
  out data_dir r17
  clr usb_wr
  set usb_wr
  ld r17 $00
  out data_dir r17
  out data_out r17
  pop r17
ret

latch_write:
  push r17
  out data_out r16
  ld r17 $FF
  out data_dir r17
  clr latch_wr
  set latch_wr
  ld r17 $00
  out data_dir r17
  out data_out r17
  pop r17
ret

delay5ms:  # 8 MHz CPU = 40,000 cycles
  push r17
  push r16
  ld r17 52; ld r16 238; .1; dec r16; jic z .1; dec r17; jic z .1
#  ld r16 128; .1; nop; nop; dec r16; jic z .1
  pop r16
  pop r17
  nop
ret
